from unittest import TestCase

from fastani.config import VERSIONS
from fastani.version import get_fastani_version


class TestVersion(TestCase):

    def test_get_version(self):
        for cur_version in VERSIONS[VERSIONS.index('1.1 or 1.2') + 1:]:
            version = get_fastani_version(f'fastANI_{cur_version}')
            self.assertEqual(version, cur_version)

        self.assertEqual(get_fastani_version(f'fastANI_1.0'), '1.0')
        self.assertEqual(get_fastani_version(f'fastANI_1.1'), '1.1 or 1.2')
        self.assertEqual(get_fastani_version(f'fastANI_1.2'), '1.1 or 1.2')



from typing import Collection, Union, FrozenSet

from fastani.logger import log


def validate_paths(paths: Union[str, Collection[str]]) -> FrozenSet[str]:
    """Validate the query/reference input."""
    out = set()
    if isinstance(paths, str):
        out.add(paths)
    elif isinstance(paths, Collection):
        for path in paths:
            if path not in out:
                out.add(path)
            else:
                log.warning(f'Duplicate path ignored: {path}')
    else:
        raise ValueError('paths must be a string or a collection of strings')
    return frozenset(out)



from fastani.model import FastANIResult

GENOMES = {'a': '/genomes/GCA_001873845.1.fna.gz',
           'b': '/genomes/GCA_002783845.1.fna.gz',
           'c': '/genomes/GCA_002789105.1.fna.gz',
           'd': '/genomes/GCA_001886815.1.fna.gz'}

KNOWN_ANI = {
    'a': {
        'b': FastANIResult(ani=99.2035, n_frag=249, total_frag=417),
        'c': FastANIResult(ani=98.6394, n_frag=225, total_frag=417),
        'd': None
    },
    'b': {
        'a': FastANIResult(ani=99.5265, n_frag=249, total_frag=282),
    },
    'c': {
        'a': FastANIResult(ani=99.4333, n_frag=224, total_frag=254),
    },
    'd': {
        'a': None,
    }
}

EXE = ('fastANI_1.0', 'fastANI_1.1', 'fastANI_1.2', 'fastANI_1.3',
       'fastANI_1.31', 'fastANI_1.32', 'fastANI_1.33')



from unittest import TestCase

from fastani.util import validate_paths


class TestValidate(TestCase):

    def test_transform_input_string(self):
        self.assertSetEqual(frozenset(['/tmp/a.txt']), validate_paths('/tmp/a.txt'))

    def test_transform_input_collection(self):
        data = ['/tmp/a.txt', '/tmp/b.txt', '/tmp/a.txt']
        self.assertSetEqual(frozenset(data), validate_paths(data))



import os
import re
import subprocess
import tempfile
from multiprocessing import Pool
from typing import Optional, Collection, Tuple, Dict

from tqdm import tqdm

from fastani.model import FastANIParameters, FastANIExecution
from fastani.model import write_genome_list

re_missing_file = re.compile(r'ERROR, skch::validateInputFiles, Could not open (.+)')


def execute_mp_wrapper(args: Tuple[FastANIParameters, Dict[str, str]]) -> FastANIExecution:
    """Required as multiprocessing pool will only allow one argument."""
    return execute(params=args[0], **args[1])


def execute(params: FastANIParameters, q: Optional[str] = None, ql: Optional[Collection[str]] = None,
            r: Optional[str] = None, rl: Optional[Collection[str]] = None) -> FastANIExecution:
    cmd = [params.exe]
    if q is not None:
        cmd.extend(['--query', q])
    if r is not None:
        cmd.extend(['--ref', r])
    cmd.extend(params.k_cmd)
    cmd.extend(params.cpu_cmd)
    cmd.extend(params.frag_len_cmd)
    cmd.extend(params.min_frac_cmd)
    cmd.extend(params.min_frag_cmd)

    with tempfile.TemporaryDirectory(prefix='fastani_') as temp_dir:

        # Set a temporary output file path if output is not set
        out_path = os.path.join(temp_dir, 'output.txt')
        cmd.extend(['--output', out_path])

        # Create a temporary file for the query/ref lists and write to disk
        if ql is not None:
            path_ql = os.path.join(temp_dir, 'query.txt')
            write_genome_list(ql, path_ql)
            cmd.extend(['--queryList', path_ql])
        if rl is not None:
            path_rl = os.path.join(temp_dir, 'reference.txt')
            write_genome_list(rl, path_rl)
            cmd.extend(['--refList', path_rl])

        # Run the proc
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf-8')
        stdout, stderr = proc.communicate()
        if proc.returncode != 0:
            missing_file = re_missing_file.search(stderr)
            if missing_file:
                raise FileNotFoundError(missing_file.group(1))
            raise Exception(f'Failed to run {params.exe} with error: {stderr}')

        # Package the results and return them
        return FastANIExecution(' '.join(cmd), stdout, stderr, proc.returncode, output=out_path)


def process_queue(params: FastANIParameters, queue) -> Tuple[FastANIExecution]:
    # If only using one CPU, don't create a multiprocessing pool
    if params.cpus == 1 or len(queue) <= 1 or params.single_execution:
        return tuple([execute(params=params, **cmd) for cmd in queue])

    # Do single executions, i.e. force FastANI execution to be single threaded
    else:
        mp_cpus = params.cpus
        params.cpus = 1
        mp_queue = [(params, item) for item in queue]
        with Pool(processes=mp_cpus) as pool:
            return tuple(tqdm(pool.imap_unordered(execute_mp_wrapper, mp_queue), total=len(mp_queue)))



__title__ = 'fastani'
__description__ = 'A wrapper to the FastANI program.'
__url__ = 'https://github.com/aaronmussig/FastANI'
__version__ = '1.1.0'
__author__ = 'Aaron Mussig'
__author_email__ = 'aaronmussig@gmail.com'
__license__ = 'GNU AGPLv3'
__bug_url__ = 'https://github.com/aaronmussig/FastANI'
__doc_url__ = 'https://github.com/aaronmussig/FastANI'
__src_url__ = 'https://github.com/aaronmussig/FastANI'

from fastani.fastani import fastani
